/*
*  Name:      application.hh
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   UESQLC application
*  Date:      $Date: 1999/08/20 11:27:01 $
*  Revision:  $Revision: 1.13 $
*
*  Copyright (C) 1996-1999  Rafael Jesus Alcantara Perez <ralcan@vnet.es>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

#ifndef _UESQLC_APPLICATION__
#define _UESQLC_APPLICATION__

#include <cstdio>
#include <mpcl/base_application.hh>
#include <mpcl/config_processor.hh>
#include <mpcl/extended_map.hh>
#include <mpcl/pattern_based_code_generator.hh>
#include <mpcl/smart_pointer.hh>
#include <string>
#include "defs.hh"
#include "requirement_verifier.hh"


//
//  U S E D   N A M E S P A C E S
//

using namespace mpcl;


///  UESQLC compiler application.
class TApplication : public TBaseApplication
{

  protected:

    ///  Smart pointer to pattern based code generator type definition.
    typedef 
      TSmartPointer <TPatternBasedCodeGenerator>  
      QTPatternBasedCodeGenerator;

    ///  Map of target run-time string to variable maps type definition.
    typedef
      xextended_map<string, XVariableMap>
      XTargetToRuntimeParametersMapMap;


  protected:

    ///  True if <help> option has been parsed in configuration stage.
    bool                                                 gCfgHelp;

    ///  True if <no line directives> option has been parsed in configuration stage.
    bool                                                 gCfgLineDirectives;

    ///  True if <output cdml> option has been parsed in configuration stage.
    bool                                                 gCfgOutputCdml;

    ///  True if <output cdml skel> option has been parsed in configuration stage.
    bool                                                 gCfgOutputCdmlSkel;

    ///  True if <parameters printing> option has been parsed in configuration stage.
    bool                                                 gCfgParameters;

    ///  True if <version> option has been parsed in configuration stage.
    bool                                                 gCfgVersion;

    ///  Compilation successfulness.
    bool                                                 gCompilationOk;

    ///  True if the current target has been loaded before.
    bool                                                 gReusingTarget;

    ///  Pointer to runtime variable map for code-generator.
    XVariableMap*                                        ptRuntimeParametersMap;

    ///  Variable map with arguments for code-generator.
    XVariableMap                                         tCodeParametersMap;

    ///  Configuration processor.
    TConfigProcessor                                     tConfigProcessor;

    ///  Requirement verifier for compiler.
    TRequirementVerifier                                 tRequirementVerifier;

    /**
    *  Map with correspondences between reserved words character
    *  representation and, their token value.
    */
    xextended_map<string, long int>                      tReservedWordMap;

    ///  Code-generators map.
    xextended_map<string, QTPatternBasedCodeGenerator>   tTargetToCodeGeneratorMap;

    ///  Map with target to runtime parameters map.
    XTargetToRuntimeParametersMapMap                     tTargetToRuntimeParametersMapMap;

    ///  Debug level from configuration stage.
    unsigned int                                         uiCfgDebug;

    ///  Path for CDML files.
    string                                               yCfgCdmlPath;

    ///  Name of the input file name.
    string                                               yCfgInputFileName;

    ///  Name of the output file name.
    string                                               yCfgOutputFileName;

    ///  Target platform.
    string                                               yCfgTarget;

    ///  Current target name.
    string                                               yCurrentTarget;

    ///  String with the last token read from scanner in symbolic form.
    string                                               yLastToken;

    ///  Output code.
    string                                               yOutputCode;

    ///  Postfix output code.
    string                                               yPostfixOutputCode;

    ///  Prefix output code.
    string                                               yPrefixOutputCode;

    ///  Line number for embedded introducer for UESQL in host language.
    size_t                                               zEmbeddedIntroducerLine;


  public:

    //
    //  C O N S T R U C T O R S
    //

    /**
    *  Builds a new instance.
    *  @param pkcAPPLICATION_NAME    Application name.
    *  @param pkcAPPLICATION_RELEASE Application release.
    */
    TApplication (const char* pkcAPPLICATION_NAME, const char* pkcAPPLICATION_RELEASE);

    ///  Destroys the instance.
    ~TApplication (void);

    /**
    *  Initialize from parameters.
    *  @param iPARAMETERS_COUNT   Main argc.
    *  @param ppkcPARAMETERS_LIST Main argv.
    */
    void initialize (int iPARAMETERS_COUNT, const char** ppkcPARAMETERS_LIST);

    /**
    *  Opens files and runs the compiler.
    *  @return Exit status.
    *  @exception TExceptionUESQLC_Fatal When any non-expected error occurs
    */
    int start (void)
      throw (TExceptionUESQLC_Fatal);

    /**
    *  Loads the code-generator with the proper CDML target file.
    *  @param pkcTARGET Target code generator.
    */
    void setCodeGenerator (const char* pkcTARGET);

    /**
    *  Initializes \a tCodeParametersMap with one value.
    *  That value is "_target", that contains the current target value.
    */
    void initCodeParameters (void);

    /**
    *  Bind code parameter \a rktVARIABLE with value \a rkyVALUE.
    *  @param rkyVARIABLE Code parameter name.
    *  @param rkyVALUE    Code parameter value.
    */
    void bindCodeParameter (const string& rkyVARIABLE, const string& rkyVALUE);

    /**
    *  Append on output, \a zLENGTH characters from \a pkcTEXT.
    *  @param pkcTEXT Text.
    *  @param zLENGTH Text length.
    */
    void outputCode (const char* pkcTEXT, size_t zLENGTH = 0);

    /**
    *  Writes at the begining of output, the code corresponding to instruction
    *  \a pkcINSTRUCTION_NAME and clause \a pkcCLAUSE_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    */
    void preAppendClauseCode (const char* pkcINSTRUCTION_NAME, const char* pkcCLAUSE_NAME = "body");

    /**
    *  Writes at the end of output, the code corresponding to instruction
    *  \a pkcINSTRUCTION_NAME and clause \a pkcCLAUSE_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    */
    void postAppendClauseCode (const char* pkcINSTRUCTION_NAME, const char* pkcCLAUSE_NAME = "body");

    /**
    *  Writes on output, the code corresponding to instruction
    *  \a pkcINSTRUCTION_NAME and clause \a pkcCLAUSE_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    */
    void outputClauseCode (const char* pkcINSTRUCTION_NAME, const char* pkcCLAUSE_NAME = "body");

    /**
    *  Sets the last read token from scanner.
    *  @param pkcLAST_TOKEN Last token read.
    */
    void pushTokenName (const char* pkcLAST_TOKEN);

    /**
    *  Sets the last read token from scanner.
    *  @param kyLAST_TOKEN Last token read.
    */
    void pushTokenName (const string& kyLAST_TOKEN);

    /**
    *  Search reserved-word map for the token binded to \a pkcTEXT, and if it
    *  is found, then the token is returned, else token \b IDENTIFIER_BODY is
    *  returned.
    *  @param pkcTEXT Token name.
    *  @return Token code.
    */
    int identifyToken (const char* pkcTEXT);

    /**
    *  Extracts target from \b EMBEDDED_SQL_BEGIN and sets the proper CDML target.
    *  @param pkcTEXT Text with parsed string.
    */
    void handleTarget (const char* pkcTEXT)
      throw (TExceptionUESQLC_Fatal, TExceptionPBCG_FileNotFound);

    /**
    *  Collapses multiple string_literal parts into an only one.
    *  @param pkcTEXT Text with multiple string_literal parts.
    *  @return Collapsed string.
    */
    string collapseStringLiteral (const char* pkcTEXT);
    

  public:

    //
    //  S E L E C T O R S
    //

    /**
    *  Returns true if current target has been used before.
    *  @return True if reusing target.
    */
    bool isReusingTarget (void) const;

    /**
    *  Returns true if compiler must use '#line' directives.
    *  @return True if using line directives.
    */
    bool useLineDirective (void) const;
    
    /**
    *  Returns compiler program file name.
    *  @return Program file name.
    */
    const string programFileName (void) const;
    
    /**
    *  Gets line directive (if any).
    *  @param gEMBEDDED_INTRODUCER_LINE_NUMBER True if line directives use the
    *                                          embedded introducer line or YACC
    *                                          variable \b yylineno.
    *  @return Text for line directive.
    */
    string getLineDirective (bool gEMBEDDED_INTRODUCER_LINE_NUMBER = false) const;

    /**
    *  Gets clause body for clause \a pkcCLAUSE_NAME in instruction
    *  \a pkcINSTRUCTION_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    *  @return Clause code.
    */
    string getClauseCode (const char* pkcINSTRUCTION_NAME, const char* pkcCLAUSE_NAME = "body");

    /**
    *  Verifies that requirement corresponding to \a liREQUERIMENT_IDENTIFIER,
    *  is true. If not, then throws the corresponding exception.
    *  @param liREQUERIMENT_IDENTIFIER Requirement identifier.
    *  \warning The proper declaration for this member function is:
    *           void require (long int, TStackValue*...);
    */
    void require (long int liREQUERIMENT_IDENTIFIER...) const;

    /**
    *  Handles syntactic errors.
    *  Throws an TExceptionUESQLC_Error exception with \a pkcMESSAGE message.
    *  @param pkcMESSAGE Error message.
    *  @exception TExceptionUESQLC_Error Exception for errors.
    */
    void handleCompilerError (const char* pkcMESSAGE) const
      throw (TExceptionUESQLC_Error);
    
    /**
    *  Create complementary files (if any). These files correspond with all
    *  variables named as "file::*".
    */
    void createComplementaryFiles (void) const
      throw (TExceptionUESQLC_Fatal);

    ///  Gets debug description, according with debug level.
    string debugDescription (void) const;

    ///  Writes to standard error a debug description, according with debug level.
    void printDebugInformation (void) const;


  protected:

    //
    //  S E L E C T O R S
    //

    ///  Writes to standard error the exception description.
    void printExceptionDescription (const TBaseException& rktSOURCE_EXCEPTION) const;

    ///  Writes to standard output the usage information.
    void printUsage (void) const;

    ///  Writes to standard output the configuration parameters.
    void printParameters (void) const;


  protected:

    //
    //  C O N S T R U C T O R S
    //

    ///  Opens files.
    void openFiles (void);

    ///  Closes files.
    void closeFiles (void);

};  // class TApplication


//
//  C O N S T R U C T O R S
//

inline void TApplication::bindCodeParameter (const string& rkyVARIABLE, const string& rkyVALUE)
{

  tCodeParametersMap.bind (rkyVARIABLE, TVariableString (rkyVALUE));

}  // bindCodeParameter()


//
//  E X P O R T E D   V A R I A B L E S
//

extern char    yytext[];
extern int     yylineno;
extern int     yydebug;
extern FILE*   yyin;
extern FILE*   yyout;


//
//  E X P O R T E D   F U N C T I O N S
//

extern int yyparse (void);


#endif  // not _UESQLC_APPLICATION__
